from bgl import *
import blf
from bge import render, logic, texture

import math

class Rectangle:
    
    def __init__(self, x=0, y=0, width=1, height=1):
        """
            Simple Rectangle class.
            
            :param x: X position
            :param y: Y position
            :param width: Width
            :param height: Height
        """
        
        # This works with Screen percentage, 0.5 is the middle of the screen
        swidth  = render.getWindowWidth()
        sheight = render.getWindowHeight()
        
        self.x      = x * swidth
        self.y      = y * sheight
        self.width  = width * swidth
        self.height = height * sheight
    
    def intersects(self, other):
        """
            Checks if this rectangle intersects with another using AABB.
            
            :param other: The other rectangle
            :return: A value indicating whether both rectangles are intersecting.
        """
        return (self.x               <= other.x + other.width  and
                self.x + self.width  >= other.x                and
                self.y               <= other.y + other.height and
                self.y + self.height >= other.y)
                
    def hasPoint(self, x, y):
        """
            Checks if this rectangle contains a point.
            
            :param x & y: The point
            :return: A value indicating whether the rectangle has a point.
        """
        return (x >= self.x              and
                x <= self.x + self.width and
                y >= self.y              and
                y <= self.y + self.height)
        
    def inflate(self, offsetx=1, offsety=1):
        """
            Increases the rectangle size by an offset value.
            
            :param offsetx: X offset
            :param offsety: Y offset
        """
        self.x      -= offsetx
        self.y      -= offsety
        self.width  += offsetx * 2
        self.height += offsety * 2

def h_gen_texture():
    id = Buffer(GL_INT, 1)
    glGenTextures(1, id)
    return id.to_list()[0]

def h_del_texture(tex):
    id = Buffer(GL_INT, 1, [tex])
    glDeleteTextures(1, id)

# Based on Moguri's BGUI
class Texture2D:
    def __init__(self, path, interpolation):
        if not hasattr(logic, "texture_cache"):
            logic.texture_cache = {}
            
        self.id = h_gen_texture()
        self.size = [0, 0]
        self.path = None
        self._interpolation = None
        self._cache = False
        
        self.bind()
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)
        self.interpolation = interpolation
        
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
        
        self.reload(path)
    
    def __del__(self):
        h_del_texture(self.id)
    
    @property
    def interpolation(self):
        return self._interpolation
    
    @interpolation.setter
    def interpolation(self, i):
        if i != self._interpolation:
            self.bind()
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, i)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, i)
            self._interpolation = i
    
    def bind(self):
        glBindTexture(GL_TEXTURE_2D, self.id)    

def setColor(col):
    glColor4f(*col)

def setWidth(w):
    glLineWidth(w)

def drawRect(x, y, width, height, mode=GL_LINE_LOOP, uv=[(0,0), (1,0), (1,1), (1,0)]):
    """
        Draw a simple rectangle using the BGL module.
        
        :param x: X position
        :param y: Y position
        :param width: Width
        :param height: Height
    """
    
    glBegin(mode)
    
    glTexCoord2f(uv[0][0], uv[0][1])
    glVertex2f(x        ,          y)
    
    glTexCoord2f(uv[1][0], uv[1][1])
    glVertex2f(x + width,          y)
    
    glTexCoord2f(uv[2][0], uv[2][1])
    glVertex2f(x + width, y + height)
    
    glTexCoord2f(uv[3][0], uv[3][1])
    glVertex2f(x        , y + height)
    
    glEnd()
    
def fillRect(x, y, width, height, uv=[(0,0), (1,0), (1,1), (1,0)]):
    """
        Draw a simple filled rectangle using the BGL module.
        
        :param x: X position
        :param y: Y position
        :param width: Width
        :param height: Height
    """
    
    drawRect(x, y, width, height, mode=GL_QUADS, uv=uv)

def drawArc(x, y, radius, start_angle, arc_angle, segs=32, mode=GL_LINE_LOOP, close=True):
    """
        Draw a simple arc.
        
        :param x: X position
        :param y: Y position
        :param radius: Arc radius
        :param start_angle: Angle to begin
        :param arc_angle: Angle to end
    """
    
    arc_angle   = arc_angle if arc_angle < math.pi*2 else math.pi*2
    start_angle = start_angle if start_angle < math.pi*2 else math.pi*2
    
    theta             = arc_angle / (segs - 1)
    tangencial_factor = math.tan(theta)
    radial_factor     = math.cos(theta)
    
    sx = math.cos(start_angle) * radius
    sy = math.sin(start_angle) * radius
        
    if close:
        glBegin(mode)
        glVertex2f(x, y)
    else:
        if mode == GL_LINE_LOOP:
            mode = GL_LINE_STRIP
        glBegin(mode)
        
    for i in range(segs):
        glVertex2f(sx + x, sy + y)
        
        tx = -sy
        ty =  sx
        
        sx += tx * tangencial_factor
        sy += ty * tangencial_factor
        
        sx *= radial_factor
        sy *= radial_factor
    glEnd()

def fillArc(x, y, radius, start_angle, arc_angle, segs=32):
    drawArc(x, y, radius, start_angle, arc_angle, segs=segs, mode=GL_TRIANGLE_FAN, close=True)

def drawLine(x1, y1, x2, y2):
    """
        Draw a simple line.
        
        :param x1 & y1: First point
        :param x2 & y2: Second point
    """
    glBegin(GL_LINES)
    
    glVertex2f(x1, y1)
    glVertex2f(x2, y2)
    
    glEnd()

def drawRoundRect(x, y, width, height, radius=2):
    """
        Draw a rounded rectangle.
        
        :param x: X position
        :param y: Y position
        :param width: Width
        :param height: Height
        :param radius: The border radius
    """
    
    if radius > 0:
        drawArc(x+radius, y+radius, radius, math.pi, math.pi*0.5, close=False)
        drawArc((x+width)-radius, y+radius, radius, (math.pi*3)*0.5, math.pi*0.5, close=False)
        drawArc((x+width)-radius, (y+height)-radius, radius, math.pi*2, math.pi*0.5, close=False)
        drawArc(x+radius, (y+height)-radius, radius, math.pi*0.5, math.pi*0.5, close=False)
    
    drawLine(x+radius, y, (x+width)-radius, y)
    drawLine(x, y+radius, x, (y+height)-radius)
    drawLine(x+radius, (y+height), (x+width)-radius, (y+height))
    drawLine(x+width, y+radius, x+width, (y+height)-radius)

def fillRoundRect(x, y, width, height, radius=2):
    """
        Draw a filled rounded rectangle.
        
        :param x: X position
        :param y: Y position
        :param width: Width
        :param height: Height
        :param radius: The border radius
    """
    
    if radius > 0:
        fillArc(x+radius, y+radius, radius, math.pi, math.pi*0.5)
        fillArc((x+width)-radius, y+radius, radius, (math.pi*3)*0.5, math.pi*0.5)
        fillArc((x+width)-radius, (y+height)-radius, radius, math.pi*2, math.pi*0.5)
        fillArc(x+radius, (y+height)-radius, radius, math.pi*0.5, math.pi*0.5)
        
        fillRect(x+radius, y, width-radius*2, radius)
        fillRect(x+radius, (y+height)-radius, width-radius*2, radius)
        fillRect(x, y+radius, radius, height-radius*2)
        fillRect((x+width)-radius, y+radius, radius, height-radius*2)
    
    fillRect(x+radius, y+radius, width-radius*2, height-radius*2)

def clipBegin(x, y, w, h, padding=[0, 0, 0, 0]):
    """
        Clips a region on the screen. Useful for text areas.
        
        :param x: X position
        :param y: Y position
        :param w: Width
        :param h: Height
        :param padding: Offset
    """
    
    vp = Buffer(GL_INT, 4)
    glGetIntegerv(GL_VIEWPORT, vp)
    
    B = [x+padding[0], y+padding[1], w-padding[2]*2, h-padding[3]*2]
    
    # Do some math to invert the coords
    scp = [0, 0, int(B[2]), int(B[3])]
    scp[0] = int(B[0] + vp[0])
    scp[1] = int(vp[1] + (vp[3] - B[1] - B[3]))
    
    glEnable(GL_SCISSOR_TEST)
    glClearColor(0.0, 0.0, 0.0, 0.0)
    glClear(GL_COLOR_BUFFER_BIT | GL_SCISSOR_BIT)
    glScissor(*scp)

def clipEnd():
    glDisable(GL_SCISSOR_TEST)

def drawText(id, x, y, width, height, size, text, vert_align, hori_align, margin, clip=True):
    """
        Draw HQ text using the BLF module.
        
        :param x: X position
        :param y: Y position
        :param width: Bounding box Width
        :param height: Bounding box Height
        :param size: Font size (percent)
        :param text: The text to be rendered
        :param vert_align: Vertical text alignment. (0 = Top, 1 = Center, 2 = Bottom)
        :param gori_align: Horizontal text alignment. (0 = Left, 1 = Center, 2 = Right)
    """
    
    if clip:
        clipBegin(x, y, width, height)
    
    glPushMatrix()
    
    rw = render.getWindowWidth()
    rh = render.getWindowHeight()
    asp = rw / rh
    
    blf.size(id, int((rh * size) / asp), 72)
    
    w, h = blf.dimensions(id, text)
    glTranslated(0.0, h, 0.0)
    glScalef(1.0, -1.0, 1.0)
    
    texts = text.split("\n")
    yn = 0
    if vert_align == 0:
        yn = margin
    elif vert_align == 1:
        yn = height / 2 - (h * len(texts)) / 2
    elif vert_align == 2:
        yn = (height - (h * len(texts))) - margin
    for i in range(len(texts)):            
        texts[i] = texts[i].replace("\t", "        ")
        wl, hl = blf.dimensions(id, texts[i])
        xn = 0
        
        if hori_align == 0:
            xn = margin
        elif hori_align == 1:
            xn = width / 2 - wl / 2
        elif hori_align == 2:
            xn = (width - wl) - margin
            
        blf.position(id, x + xn, -y - (i * h) - yn, 1)
        
        blf.draw(id, texts[i])
        
    glScalef(1.0, 1.0, 1.0)
    
    glPopMatrix()
    
    if clip:
        clipEnd()

def setMode2D():
    """
        Set a 2D rendering mode
    """
    
    width  = render.getWindowWidth()
    height = render.getWindowHeight()
    
    # 2D Projection
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    glOrtho(0, width, height, 0, -1, 1)
    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()
    
    # 2D Shading
    glDisable(GL_CULL_FACE)
    glDisable(GL_LIGHTING)
    glDisable(GL_DEPTH_TEST)
    glShadeModel(GL_SMOOTH)
    
    # Line antialias
    glEnable(GL_LINE_SMOOTH)
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST)
    
    # 2D Blending (Alpha)
    glEnable(GL_BLEND)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

def drawTexture(id, w, h, bounds, coords, debug=False, coordpos=[0,0], invert=False):
    """
        Draw a simple texture.
        
        :param id: Texture ID
        :param w: Texture Width
        :param h: Texture Height
        :param bounds: Canvas bounds
        :param coords: Texture UV
    """
    
    glEnable(GL_TEXTURE_2D)
    glBindTexture(GL_TEXTURE_2D, id)
    glEnable(GL_BLEND)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
    
    glColor4f(*(1,1,1,1))
    
    B = bounds
    
    if invert:
        C = [
            (coords[0][0]+coordpos[0], coords[2][1]+coordpos[1]),
            (coords[1][0]+coordpos[0], coords[3][1]+coordpos[1]),
            (coords[2][0]+coordpos[0], coords[0][1]+coordpos[1]),
            (coords[3][0]+coordpos[0], coords[1][1]+coordpos[1])
        ]
    else:
        C = [
            (coords[0][0]+coordpos[0], coords[0][1]+coordpos[1]),
            (coords[1][0]+coordpos[0], coords[1][1]+coordpos[1]),
            (coords[2][0]+coordpos[0], coords[2][1]+coordpos[1]),
            (coords[3][0]+coordpos[0], coords[3][1]+coordpos[1])
        ]
        
    fillRect(B[0], B[1], B[2], B[3], uv=C)
    
    if debug:
        glBindTexture(GL_TEXTURE_2D, 0)
        glColor4f(*(0,1,0,1))
        drawRect(int(B[0]), int(B[1]), int(B[2]), int(B[3]))

    glBindTexture(GL_TEXTURE_2D, 0)
    glDisable(GL_TEXTURE_2D)
    
def drawNinePatch(id, w, h, bounds, padding, addcoordspos=[0,0], drawcenter=True):
    B = bounds
    P = padding
    A = addcoordspos
    
    x = (B[0] / w)
    y = (B[1] / h)
    
    # LEFT TOP
    uv_pad_left = (P[0] / w)
    uv_pad_top = (P[1] / h)
    
    C = [
        (0          , 0),
        (uv_pad_left, 0), 
        (uv_pad_left, uv_pad_top),
        (0          , uv_pad_top)
    ]    
    drawTexture(id, w, h, [B[0], B[1], P[0], P[1]], C, coordpos=A)
    
    # MIDDLE TOP
    uv_pad_right = 1.0 - (P[2] / w)
    
    C = [
        (uv_pad_left , 0),
        (uv_pad_right, 0), 
        (uv_pad_right, uv_pad_top),
        (uv_pad_left , uv_pad_top)
    ]
    drawTexture(id, w, h, [B[0]+P[0], B[1], B[2]-(P[2]+P[0]), P[1]], C, coordpos=A)
    
    # RIGHT TOP
    uv_pad_right2 = uv_pad_right + (P[2] / w)
    C = [
        (uv_pad_right , 0),
        (uv_pad_right2, 0), 
        (uv_pad_right2, uv_pad_top),
        (uv_pad_right , uv_pad_top)
    ]
    drawTexture(id, w, h, [B[0]+B[2]-P[2], B[1], P[2], P[1]], C, coordpos=A)
    
    # LEFT MIDDLE
    uv_pad_bottom = 1.0 - (P[3] / w)
    C = [
        (0          , uv_pad_top),
        (uv_pad_left, uv_pad_top), 
        (uv_pad_left, uv_pad_bottom),
        (0          , uv_pad_bottom)
    ]
    drawTexture(id, w, h, [B[0], B[1]+P[1], P[0], B[3]-(P[3]+P[1])], C, coordpos=A)
    
    # CENTER MIDDLE
    if drawcenter:
        C = [
            (uv_pad_left , uv_pad_top),
            (uv_pad_right, uv_pad_top), 
            (uv_pad_right, uv_pad_bottom),
            (uv_pad_left , uv_pad_bottom)
        ]
        drawTexture(id, w, h, [B[0]+P[0], B[1]+P[1], B[2]-(P[2]+P[0]), B[3]-(P[3]+P[1])], C, coordpos=A)
    
    # RIGHT MIDDLE
    C = [
        (uv_pad_right , uv_pad_top),
        (uv_pad_right2, uv_pad_top), 
        (uv_pad_right2, uv_pad_bottom),
        (uv_pad_right , uv_pad_bottom)
    ]
    drawTexture(id, w, h, [B[0]+B[2]-P[2], B[1]+P[1], P[2], B[3]-(P[3]+P[1])], C, coordpos=A)
    
    # LEFT BOTTOM
    uv_pad_bottom2 = uv_pad_bottom + (P[3] / h)
    C = [
        (0          , uv_pad_bottom),
        (uv_pad_left, uv_pad_bottom), 
        (uv_pad_left, uv_pad_bottom2),
        (0          , uv_pad_bottom2)
    ]
    drawTexture(id, w, h, [B[0], B[1]+(B[3]-P[3]), P[0], P[3]], C, coordpos=A)
    
    # MIDDLE BOTTOM
    C = [
        (uv_pad_left , uv_pad_bottom),
        (uv_pad_right, uv_pad_bottom), 
        (uv_pad_right, uv_pad_bottom2),
        (uv_pad_left , uv_pad_bottom2)
    ]
    drawTexture(id, w, h, [B[0]+P[0], B[1]+(B[3]-P[3]), B[2]-(P[0]+P[2]), P[3]], C, coordpos=A)
    
    # RIGHT BOTTOM
    C = [
        (uv_pad_right , uv_pad_bottom),
        (uv_pad_right2, uv_pad_bottom), 
        (uv_pad_right2, uv_pad_bottom2),
        (uv_pad_right , uv_pad_bottom2)
    ]
    drawTexture(id, w, h, [B[0]+B[2]-P[2], B[1]+(B[3]-P[3]), P[2], P[3]], C, coordpos=A)


class Image(Texture2D):
    def __init__(self, path, interp=GL_LINEAR, cache=True):
        Texture2D.__init__(self, path, interp)
        self._cache = cache
        self.valid = True
    
    def draw(self, x, y, w, h):
        drawTexture(self.id, self.size[0], self.size[1], [x,y,w,h], [(0,0),(1,0),(1,1),(0,1)])
    
    def reload(self, path):
        glEnable(GL_TEXTURE_2D)
        
        if path == self.path: return
        if path in logic.texture_cache:
            img = logic.texture_cache[path]
        else:
            img = texture.ImageFFmpeg(path)
            img.scale = False
            if self._cache:
                logic.texture_cache[path] = img
        
        data = img.image
        if data == None:
            print("Could not load the image", img)
            self.valid = False
            return
                
        self.bind()
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, img.size[0],
                         img.size[1], 0, GL_RGBA,
                         GL_UNSIGNED_BYTE, data)
        
        self.size = img.size[:]
        self.path = path
        
        img = None