from bge import logic as g, events as e, render
from bgl import *
import blf
#import kxg_util as Util
from .kxg_util import *

g.current_focus = None
g.handled = False

def k_down(btn, state=g.KX_INPUT_ACTIVE):
    return g.keyboard.events[btn] == state
def k_pressed(btn, state=g.KX_INPUT_JUST_ACTIVATED):
    return g.keyboard.events[btn] == state
def k_released(btn, state=g.KX_INPUT_JUST_RELEASED):
    return g.keyboard.events[btn] == state

def get_all_keys():
    keys = {}
    for k, v in e.__dict__.items():
        if not k.startswith("__"):
            if k not in ["EventToCharacter", "EventToString",
                         "RETKEY", "WHEELUPMOUSE",
                         "WHEELDOWNMOUSE", "MOUSEY", "MOUSEX",
                         "LEFTMOUSE", "MIDDLEMOUSE", "RIGHTMOUSE"]:
                keys[k] = v
    return keys

supported_keys = get_all_keys()

class Group:
    
    def __init__(self, position=[0,0]):
        self.visible = True
        self.active = True
        self.position = position
        
        self.stack = True
        self.stackSpacing = 5
        
        self.widgets = []
    
    @property
    def position(self):
        return self._position
    
    @position.setter
    def position(self, pos):
        width  = render.getWindowWidth()
        height = render.getWindowHeight()
        self._position = [pos[0] * width, pos[1] * height]
    
    def onUpdate(self):
        for widget in self.widgets:
            widget._bounds.x = widget._obounds.x + self.position[0]
            widget._bounds.y = widget._obounds.y + self.position[1]
            if self.stack:
                widget._bounds.y += (widget.bounds.height + self.stackSpacing) * widget.index

class Patch9(Image):
    
    def __init__(self, image_path="", interp=GL_NEAREST):
        Image.__init__(self, image_path, interp=interp)
        
        self.padding = [5, 5, 5, 5]
        self.drawCenter = True
        self.uvOffset = [0, 0]
    
    def draw(self, x, y, w, h):
        drawNinePatch(self.id,
                           self.size[0], self.size[1],
                           [x, y, w, h],
                           self.padding,
                           addcoordspos=self.uvOffset, drawcenter=self.drawCenter)

class GUI:
    
    def __init__(self):
        self.groups = {}
        self.widgets = {}
        self.cursor = None
        self.windex = 0
        
        self.theme = {}
    
    def loadTheme(self, filepath):
        import json
        
        """
            Theme structure
            
            {
                "Button": {
                    "normal": <tuple 4 or path>,
                    "click": <tuple 4 or path>,
                    "hover": <tuple 4 or path>,
                    "textColor": <tuple 4>,
                    "borderColor": <tuple 4>,
                    "borderWidth": <int>,
                    "borderRadius": <int>
                }
            }
        """
        
        ob = None
        with open(filepath, "r") as f:
            ob = json.loads(f.read())
        obk = ob.keys()
        
        if ob is not None:
            for mainkey in obk:
                for inkey in ob[mainkey].keys():
                    item = ob[mainkey][inkey]
                    if isinstance(item, dict):
                        path = g.expandPath(item["image"])
                        ob[mainkey][inkey] = Patch9(image_path=path)
                        ob[mainkey][inkey].padding = item["padding"]
                        
                self.theme[mainkey] = ob[mainkey]
    
    def applyTheme(self):
        for k, v in self.widgets.items():
            try:
                v.theme = self.theme[v.__class__.__name__]
            except:
                continue
    
    def applyToScene(self, scene=None):
        if scene is None:
            scene = g.getCurrentScene()
        scene.post_draw.append(self.draw)
    
    def createGroup(self, name):
        self.groups[name] = Group()
    
    def addToGroup(self, wname, gname):
        if not gname in self.groups.keys(): return
        if not wname in self.widgets.keys(): return
        
        self.widgets[wname].group = self.groups[gname]
        self.groups[gname].widgets.append(self.widgets[wname])
    
    def addWidget(self, name, w, group=None):
        # WARNING: Widgets with the same name will be overwritten!
        cnt = len(self.widgets.keys())
        
        w.index = self.windex
        self.widgets[name] = w
        
        if cnt <= 0:
            self.widgets[name].requestFocus()

        self.windex += 1
        if self.windex >= 999999:
            self.windex = -1
            
        if group is not None:
            if group not in self.groups.keys():
                self.createGroup(group)
            self.addToGroup(name, group)
    
    def draw(self):        
        setMode2D()
        setWidth(1)
        
        wids = sorted(self.widgets.values(), key=lambda x: x.zorder)
        for widget in wids:
            if widget.visible:
                if widget.group is not None:
                    if not widget.group.visible:
                        break
                widget.onDraw()
        
        if self.cursor != None:
            width  = render.getWindowWidth()
            height = render.getWindowHeight()
            
            mx = int(g.mouse.position[0] * width)
            my = int(g.mouse.position[1] * height)
        
            render.showMouse(0)
            self.cursor.onDraw()
        else:
            render.showMouse(1)

    def update(self):
        width  = render.getWindowWidth()
        height = render.getWindowHeight()
           
        mx = int(g.mouse.position[0] * width)
        my = int(g.mouse.position[1] * height)
        
        if self.cursor != None:
            self.cursor.position = [mx, my]
            hx, hy = self.cursor.hotSpot
        else:
            hx, hy = 0, 0

        wids = sorted(self.widgets.values(), key=lambda x: x.zorder)
        for w in wids:
            w.onUpdate(mx + hx, my + hy)

class Cursor(Image):
    
    def __init__(self, image_path="", interp=GL_NEAREST, scale=[1, 1]):
        Image.__init__(self, image_path, interp=interp)
        self.scale = scale
        self.debug = False
        self.hotSpot = [0, 0]
        self.position = [0, 0]
    
    def fromJSON(self, jsonFile):
        import json
        """
            JSON Struct
            {
                "image": "cursor1.png",
                "hotSpot": { "x": 0, "y": 0 },
                "scale": { "x": 1, "y": 1 }
            }
            
            == HotSpot Value ==================
            
            0, 0
            |
            +------------+ - 1, 0
            |            |
            |            |
            |    0.5, 0.5|
            |      .     |
            |            |
            |            |
            +------------+ - 1, 1
            |
            0, 1
            
        """
        
        with open(jsonFile, "r") as f:
            ob = json.loads(f.read())
        obk = ob.keys()
        
        if "image" in obk:
            self.reload(g.expandPath(ob["image"]))
        if "hotSpot" in obk:
            self.hotSpot = [ob["hotSpot"]["x"], ob["hotSpot"]["y"]]
        if "scale" in obk:
            self.scale = [ob["scale"]["x"], ob["scale"]["y"]]
        
    def onDraw(self):
        w, h = self.size
        id = self.id

        textureCoords = [
            (0, 1),
            (1, 1),
            (1, 0),
            (0, 0)
        ]
        
        # 0.5 is the middle
        hspot = [self.hotSpot[0] * self.size[0], self.hotSpot[1] * self.size[1]]
        
        x, y = self.position
        hx, hy = x + hspot[0], y + hspot[1]
        
        bounds = [x, y, self.scale[0] * w, self.scale[1] * h]
        if id > -1:
            drawTexture(id, w, h, bounds, textureCoords)
        
        if self.debug:
            glEnable(GL_LINE_STIPPLE);
            
            glLineStipple(2, 0xAAAA)
            setColor((0.0, 1.0, 1.0, 1))
            drawRect(x, y, w, h)
            
            glDisable(GL_LINE_STIPPLE);
            
            setColor((1.0, 1.0, 0.0, 1))
            glPointSize(3)
            glBegin(GL_POINTS)
            glVertex2f(hx, hy)
            glEnd()
            
            glBegin(GL_LINES)
            glVertex2f(hx-4, hy)
            glVertex2f(hx+4, hy)
            glVertex2f(hx, hy-4)
            glVertex2f(hx, hy+4)
            glEnd()

class Clock:
    def __init__(self):
        self.enabled = True
        self.time = 0.0
        self.delay = 0.0
        
    def reset(self):
        self.time = 0.0
    
    def update(self):
        if self.enabled:
            self.time += (1.0 / g.getLogicTicRate()) * (self.delay + 1.0)

class Events:
    
    def __init__(self, o):
        self.owner = o
        
        self.callbacks = {
            "on_click"      : None,
            "on_release"    : None,
            "on_move"       : None,
            "on_keytyped"   : None,
            "on_keyreleased": None
        }
        
        self.clicked = False
        self.hovered = False
        
        self.position = [0, 0]
        
        self.keyTyped = False
        self.key = -1
        self.ctrl = False
        self.alt = False
        self.shift = False
    
    def activate(self, name, args=[]):
        if name in self.callbacks.keys():
            cb = self.callbacks[name]
            default_args = []
            default_args.append(self)
            margs = []
            
            if len(args) > 0:
                margs = default_args.extend(args)
            else:
                margs = default_args
                
            if cb != None:
                cb(*margs)
                
                return True
        return False
    
    def register(self, name, cbk):
        self.callbacks[name] = cbk
    
class Widget:
    
    def __init__(self, x=0, y=0, width=0.2, height=0.08):
        self.bounds = Rectangle(x, y, width, height)
        
        self.index = 0
        self.group = None
        self.borderRadius = 0
        self.font = 0 # Font ID from the resource manager
        self.active = True
        self.visible = True
        self.focused = False
        self.zorder = 0
        self.drawBackground = True
        
        self.events = Events(self)
        
        self.theme = {
            "textColor": (0, 0, 0, 1),
            "backColor": (0.62, 0.61, 0.6, 1)
        }
    
    @property
    def bounds(self):
        return self._bounds
    
    @bounds.setter
    def bounds(self, b):
        import copy
        
        self._obounds = copy.copy(b)
        self._bounds = b
    
    def requestFocus(self):
        if g.current_focus != None:
            g.current_focus.focused = False
            
        g.current_focus = self
        self.focused = True
        g.handled = True
    
    def onKeyPress(self, sender):
        pass
    
    def onMouseClick(self, sender):
        pass
              
    def onDraw(self):
        rec = self.bounds
        
        setColor(self.theme["backColor"])
        fillRoundRect(rec.x, rec.y, rec.width, rec.height, radius=self.borderRadius)

    def onUpdate(self, mx, my):
        global supported_keys
        
        if self.focused:
            self.zorder = 99
        else:
            self.zorder = -1
            
        self.events.position = [
            mx - self.bounds.x,
            my - self.bounds.y
        ]
        
        if self.focused:
            self.events.keyTyped = False
                   
            self.events.shift = k_down(e.LEFTSHIFTKEY) or k_down(e.RIGHTSHIFTKEY)
            self.events.ctrl  = k_down(e.LEFTCTRLKEY)  or k_down(e.RIGHTCTRLKEY)
            self.events.alt   = k_down(e.LEFTALTKEY)   or k_down(e.RIGHTALTKEY)
            
            for k, v in supported_keys.items():
                if k_pressed(v):
                    self.events.key = v
                    self.events.keyTyped = True
                    self.events.activate("on_keytyped")
                    self.onKeyPress(self.events)
                elif k_released(v):
                    self.events.key = -1
                    self.events.keyTyped = False
                    self.events.activate("on_keyreleased")
                else: continue
            
        if self.bounds.hasPoint(mx, my):
            self.events.hovered = True
            
            self.events.activate("on_move")
            
            if not g.handled:
                if g.mouse.events[e.LEFTMOUSE] == g.KX_INPUT_ACTIVE:
                    self.events.clicked = True
                    self.events.activate("on_click")
                    self.onMouseClick(self.events)
                    self.requestFocus()
                    g.handled = True
                    
            if g.mouse.events[e.LEFTMOUSE] == g.KX_INPUT_JUST_RELEASED:
                self.events.clicked = False
                self.events.activate("on_release")
                g.handled = False
        else:
            self.events.hovered = False
            self.events.clicked = False

        if self.group is not None:
            if not self.group.active:
                return
            self.group.onUpdate()
            
class Label(Widget):
    
    def __init__(self, x=0, y=0, width=0.2, height=0.08, text="", size=0.05):
        Widget.__init__(self, x, y, width, height)
        
        self.textSize = size
        self.text = text
        self.verticalAlign = 0
        self.HorizontalAlign = 0
        self.Margin = 1
        
        self.theme = {
            "textColor": (1, 1, 1, 1),
            "backColor": (0, 0, 0, 0)
        }
        
    def onDraw(self):
        Widget.onDraw(self)
        
        setColor(self.theme["textColor"])
        drawText(self.font,
                      self.bounds.x,
                      self.bounds.y,
                      self.bounds.width,
                      self.bounds.height,
                      self.textSize,
                      self.text,
                      self.verticalAlign,
                      self.HorizontalAlign,
                      self.Margin)

class Button(Widget):
    
    def __init__(self, x=0, y=0, width=0.2, height=0.08, text="Button", size=0.05):
        Widget.__init__(self, x, y, width, height)
        
        self.textSize = size
        self.text = text
        self.verticalAlign = 1
        self.HorizontalAlign = 1
        self.Margin = 1   
        
        self.theme = {
            "textColor": (0, 0, 0, 1),
            "backColor": (0.62, 0.61, 0.6, 1),
            
            # Can be an Image or a Patch9
            "normal": (0.62, 0.61, 0.6, 1),
            "hover": (0.72, 0.71, 0.7, 1),
            "click": (0.42, 0.41, 0.4, 1),
            
            "borderColor": (0, 0, 0, 1),
            "borderWidth": 1,
            "borderRadius": 0
        }
        
    def onUpdate(self, mx, my):
        Widget.onUpdate(self, mx, my)
    
    def onDraw(self):
        evt = self.events
        b = self.bounds
        
        isImageBackground = (isinstance(self.theme["normal"], Image) and
                             isinstance(self.theme["hover"], Image) and
                             isinstance(self.theme["click"], Image))
        if not isImageBackground:
            Widget.onDraw(self)
            
        if self.drawBackground:
            if not evt.hovered and not evt.clicked:
                if not isImageBackground:
                    setColor(self.theme["normal"])
                    fillRoundRect(b.x, b.y, b.width, b.height, radius=self.theme["borderRadius"])
                else:
                    self.theme["normal"].draw(b.x, b.y, b.width, b.height)
                    
            elif evt.hovered and not evt.clicked:
                if not isImageBackground:
                    setColor(self.theme["hover"])
                    fillRoundRect(b.x, b.y, b.width, b.height, radius=self.theme["borderRadius"])
                else:
                    self.theme["hover"].draw(b.x, b.y, b.width, b.height)
            elif evt.hovered and evt.clicked:
                if not isImageBackground:
                    setColor(self.theme["click"])
                    fillRoundRect(b.x, b.y, b.width, b.height, radius=self.theme["borderRadius"])
                else:
                    self.theme["click"].draw(b.x, b.y, b.width, b.height)
            else:
                if not isImageBackground:
                    setColor(self.theme["click"])
                    fillRoundRect(b.x, b.y, b.width, b.height, radius=self.theme["borderRadius"])
                else:
                    self.theme["click"].draw(b.x, b.y, b.width, b.height)
            
            if not isImageBackground:
                setWidth(self.theme["borderWidth"])
                setColor(self.theme["borderColor"])
                drawRoundRect(b.x, b.y, b.width, b.height, radius=self.theme["borderRadius"])
                
                setWidth(1)
        
        setColor(self.theme["textColor"])
        drawText(self.font,
                      self.bounds.x,
                      self.bounds.y,
                      self.bounds.width,
                      self.bounds.height,
                      self.textSize,
                      self.text,
                      self.verticalAlign,
                      self.HorizontalAlign,
                      self.Margin)

class TextBox(Label):
    
    def __init__(self, x=0, y=0, width=0.2, height=0.2, text="", size=0.05):
        Label.__init__(self, x, y, width, height, text, size)

        self.lines = []
        self.text = text
        
        self.x = 0
        self.y = 0
        self.cx = 0
        self.cy = 0
        
        self.multiLine = True
        self.readOnly    = False
        self.charSpacing = 1
        self.lineHeight = 12
        self.lineSpacing = 1
        self.numbersOnly = False
        self.masked      = False
        self.mask = "*"
        self.blink = False
        
        self.theme = {
            "textColor": (0, 0, 0, 1),
            "backColor": (1, 1, 1, 1),
            
            # Can be a Patch9
            "normal": (0.62, 0.61, 0.6, 1),
            "hover": (0.72, 0.71, 0.7, 1),
            "click": (0.42, 0.41, 0.4, 1),
            
            "borderColor": (0, 0, 0, 1),
            "borderWidth": 1,
            "borderRadius": 0
        }
        
        self.fw, self.fh = blf.dimensions(self.font, "Ee|{^")
        
        self.clk = Clock()
        
    @property
    def text(self):
        return self._text
    
    @text.setter
    def text(self, nt):
        if nt == None:
            self._text = ""
            self.lines = [nt]
            self.x = 0
        else:
            self._text = nt
            if "\n" in nt:
                self.lines = []
                for t in nt.split("\n"):
                    self.lines.append(t)
            else:
                self.lines = [nt]
    
    def updateText(self):
        self._text = "\n".join(self.lines)
    
    def onUpdate(self, x, y):
        Widget.onUpdate(self, x, y)
        
        self.clk.update()
        
        if self.clk.time >= 0.5:
            self.clk.reset()
            self.blink = not self.blink
    
    def onMouseClick(self, sender):
        x, y = sender.position
        offx = 0
        offy = 0
        
        for j in range(len(self.lines)):
            for i in range(len(self.lines[j])):
                char = self.lines[j][i]
                charw, _ = blf.dimensions(self.font, char)
                if self.masked:
                    charw, _ = blf.dimensions(self.font, self.mask)
                
                rect = Rectangle(0,0,0,0)
                rect.x = offx
                rect.y = offy
                rect.width = charw
                rect.height = self.lineHeight
                
                if rect.hasPoint(x, y):
                    self.x = i
                    self.y = j
                    break

                offx += charw + self.charSpacing
            offy += self.lineHeight + self.lineSpacing
            offx = 0
            
    def onKeyPress(self, sender):
        if self.readOnly: return
        
        key, shift = sender.key, sender.shift        
        char = e.EventToCharacter(key, shift)
        
        ln = self.lines[self.y]
        
        if self.numbersOnly:
            if key not in [e.ONEKEY, e.TWOKEY, e.THREEKEY, e.FOURKEY, e.FIVEKEY,
                e.SIXKEY, e.SEVENKEY, e.EIGHTKEY, e.NINEKEY, e.ZEROKEY, e.PAD0,
                e.PAD1, e.PAD2, e.PAD3, e.PAD4, e.PAD5, e.PAD6, e.PAD7, 
                e.PAD8, e.PAD9, e.PADPERIOD, e.PADMINUS, e.PERIODKEY, e.MINUSKEY,
                e.LEFTARROWKEY, e.RIGHTARROWKEY, e.BACKSPACEKEY, e.DELKEY]:
                return
        
        if key == e.LEFTARROWKEY:
            if self.x > 0:
                self.x -= 1
            else:
                self.x = 0
        elif key == e.RIGHTARROWKEY:            
            if self.x < len(ln):
                self.x += 1
            else:
                self.x = len(ln)
        elif key == e.UPARROWKEY:            
            if self.y > 0:
                self.y -= 1
                if self.x > len(self.lines[self.y]):
                    self.x = len(self.lines[self.y])
            else:
                self.y = 0
        elif key == e.DOWNARROWKEY:            
            if self.y < len(self.lines)-1:
                self.y += 1
                if self.x > len(self.lines[self.y]):
                    self.x = len(self.lines[self.y])
            else:
                self.y = len(self.lines)-1
        elif key == e.TABKEY:
            ln = ln[:self.x] + "    " + ln[self.x:]
            self.lines[self.y] = ln
            self.x += 4
            self.updateText()
        elif key == e.BACKSPACEKEY:
            if self.x > 0:
                if self.x <= len(ln):
                    self.x -= 1                
                    ln = ln[:self.x] + ln[self.x+1:]
                else:
                    ln = ln[:self.x-1] + ln[self.x:]
                self.lines[self.y] = ln
                self.updateText()
            else:
                if self.y > 0:
                    self.x = len(self.lines[self.y-1])
                    self.lines[self.y-1] += self.lines[self.y]
                    del self.lines[self.y]
                    self.y -= 1
                else:
                    self.x = 0
                self.updateText()
        elif key == e.DELKEY:
            stln = ln[self.x:]            
            if len(stln) > 0:
                ln = ln[:self.x] + ln[self.x+1:]
                self.text = ln
            else:
                if len(self.lines) > 0:
                    if self.y < len(self.lines):
                        self.lines[self.y] += self.lines[self.y+1]
                        del self.lines[self.y+1]
                        self.updateText()
        elif key == e.ENTERKEY:
            if not self.multiLine: return
            bln = ln[self.x:]
            self.lines.insert(self.y+1, bln)
            self.lines[self.y] = ln[:self.x]
            self.y += 1
            self.x = 0
        elif key == e.LEFTSHIFTKEY or key == e.RIGHTSHIFTKEY:
            pass
        else:
            if self.x >= len(ln):
                ln += char
            else:
                ln = ln[:self.x] + char + ln[self.x:]
                
            self.x += 1
                
            ln = ln.replace("\n", "")
            self.lines[self.y] = ln
            self.updateText()
        
        self.blink = True
        self.__update_caret()
        
    def __update_caret(self):
        self.cx = 0
        for i in range(self.x):
            self.cx += self.fw + self.charSpacing
    
    def onDraw(self):
        import copy
        
        offx = 0
        offy = 0
        
        isImageBackground = isinstance(self.theme["normal"], Image)

        if self.drawBackground:
            if not isImageBackground:                
                setColor(self.theme["backColor"])
                
                bounds = copy.copy(self.bounds)
                bounds.inflate(2, 2)
                
                fillRoundRect(bounds.x, bounds.y,
                                   bounds.width, bounds.height,
                                   radius=self.theme["borderRadius"])
                
                setColor(self.theme["borderColor"])
                setWidth(self.theme["borderWidth"])
                drawRoundRect(bounds.x, bounds.y,
                                   bounds.width, bounds.height,
                                   radius=self.theme["borderRadius"])
            else:
                self.theme["normal"].draw(self.bounds.x, self.bounds.y,
                                         self.bounds.width, self.bounds.height)
                
        clipBegin(self.bounds.x, self.bounds.y, self.bounds.width, self.bounds.height)
        setColor(self.theme["textColor"])
        for j in range(len(self.lines)):
            for i in range(len(self.lines[j])):
                char = self.lines[j][i]
                charw, _ = blf.dimensions(self.font, char)
                
                if not self.masked:
                    drawText(self.font,
                                  self.bounds.x+offx+2,
                                  self.bounds.y+offy,
                                  charw, self.lineHeight,
                                  self.textSize,
                                  char,
                                  2, 0, 0, clip=False)
                else:
                    charw, _ = blf.dimensions(self.font, self.mask)
                    drawText(self.font,
                                  self.bounds.x+offx+2,
                                  self.bounds.y+offy,
                                  charw, self.lineHeight,
                                  self.textSize,
                                  self.mask,
                                  2, 0, 0, clip=False)
                
                offx += charw + self.charSpacing
                #setColor((1,0,0,1))
                #drawRect(self.bounds.x+offx,
                #              self.bounds.y+offy,
                #              charw, self.lineHeight)
                              
            offy += self.lineHeight + self.lineSpacing
            offx = 0
        
        if self.focused and not self.readOnly and self.blink:
            offx = 0
            offy = self.y * (self.lineHeight+self.lineSpacing)
            txt = self.lines[self.y]+" "
            
            for i in range(len(txt)):
                charw, charh = blf.dimensions(self.font, txt[i])
                if self.masked:
                    charw, charh = blf.dimensions(self.font, self.mask)
                
                if i == self.x:
                    cx = self.bounds.x + offx - 2
                    cy = self.bounds.y + offy
                    drawText(self.font,
                                  cx+2, cy,
                                  1, self.lineHeight,
                                  self.textSize,
                                  "|",
                                  2, 0, 0, clip=False)
                    break
                offx += charw + self.charSpacing
        clipEnd()
        
class ImageView(Widget):
    
    def __init__(self, x=0, y=0, width=0.2, height=0.2, image_path=""):
        Widget.__init__(self, x, y, width, height)
        
        self._texture = None
        self._src = ""
        
        self.src = image_path
        
        self.textureCoords = [
            (0, 1),
            (1, 1),
            (1, 0),
            (0, 0)
        ]
        
        self.theme = {
            "textColor": (0, 0, 0, 1),
            "backColor": (0.62, 0.61, 0.6, 1),
            "borderColor": (0, 0, 0, 1),
            "borderWidth": 1,
            "borderRadius": 0
        }
    
    @property
    def src(self):
        return self._src
    
    @src.setter
    def src(self, v):
        self._src = v
        if v != "":
            self._texture = Image(v)
    
    def onDraw(self):
        if self.drawBackground:
            b = self.bounds
            
            setWidth(self.theme["borderWidth"])
            setColor(self.theme["borderColor"])
            drawRoundRect(b.x, b.y, b.width, b.height,
                               radius=self.theme["borderRadius"])
            
            Widget.onDraw(self)
        
        if self._texture == None: return
        
        self.textureCoords = [
            (0, self._texture.size[1]),
            (self._texture.size[0], self._texture.size[1]),
            (self._texture.size[0], 0),
            (0, 0)
        ]
        
        clipBegin(self.bounds.x, self.bounds.y, self.bounds.width, self.bounds.height)
        drawTexture(self._texture.id,
                         self._texture.size[0],
                         self._texture.size[1],
                         [self.bounds.x, self.bounds.y,
                          self.bounds.width, self.bounds.height],
                          self.textureCoords)
        clipEnd()
